from zope.component import getUtility
from seantis.reservation.session import ILibresUtility


def dbsetup(context):
    scheduler = getUtility(ILibresUtility).scheduler('maintenance', 'UTC')
    scheduler.setup_database()



from datetime import datetime, timedelta

from five import grok
from zope.interface import Interface

from sqlalchemy import desc

from seantis.reservation import Session
from seantis.reservation import _
from seantis.reservation import utils
from seantis.reservation.reservations import combine_reservations
from libres.db.models import Reservation
from seantis.reservation.base import BaseView
from seantis.reservation.reports import GeneralReportParametersMixin


def human_date(date):
    # timezones are currently naive and implicity the one used by
    # the users - we don't have international reservations yet
    now = utils.utcnow()

    this_morning = datetime(
        now.year, now.month, now.day
    ).replace(tzinfo=now.tzinfo)

    time = utils.localize_date(date, time_only=True)

    if date >= this_morning:
        return _(u'Today, at ${time}', mapping={'time': time})

    days = (now.date() - date.date()).days

    if days <= 1:
        return _(u'Yesterday, at ${time}', mapping={
            'time': time
        })
    else:
        return _(u'${days} days ago, at ${time}', mapping={
            'days': days,
            'time': time
        })


class LatestReservationsReportView(BaseView, GeneralReportParametersMixin):

    permission = 'seantis.reservation.ViewReservations'

    grok.require(permission)

    grok.context(Interface)
    grok.name('latest_reservations')

    template = grok.PageTemplateFile('../templates/latest_reservations.pt')

    @property
    def title(self):
        return _(u'Latest Reservations')

    @property
    def start(self):
        return utils.safe_parse_int(self.request.get('start'), 0)

    @property
    def end(self):
        return utils.safe_parse_int(self.request.get('end'), 30)

    @property
    def results(self):
        return latest_reservations(
            resources=self.resources,
            reservations=self.reservations or '*',
            daterange=self.daterange,
        )

    @property
    def daterange(self):
        now = utils.utcnow()

        since = now - timedelta(days=self.end)
        until = now - timedelta(days=self.start)

        return since, until

    @property
    def daterange_label(self):
        since, until = self.daterange

        if until.date() == utils.utcnow().date():
            return ' - '.join((
                utils.localize_date(since, long_format=False),
                self.translate(_(u'Today'))
            ))
        return ' - '.join((
            utils.localize_date(since, long_format=False),
            utils.localize_date(until, long_format=False)
        ))

    def build_url(self, start, end):
        params = [
            ('start', str(start)),
            ('end', str(end))
        ]

        return super(LatestReservationsReportView, self).build_url(
            extra_parameters=params
        )

    @property
    def forward_url(self):
        start, end = self.start, self.end
        start, end = start - 30, end - 30

        if start < 0:
            return None

        return self.build_url(start, end)

    @property
    def backward_url(self):
        start, end = self.start, self.end
        start, end = start + 30, end + 30

        return self.build_url(start, end)

    def reservation_title(self, reservation):
        human_date_text = utils.translate(
            self.context, self.request, human_date(reservation.created)
        )
        return '{} - {}'.format(human_date_text, reservation.title)

    def unique(self, reservations):
        return tuple(combine_reservations(reservations))


def latest_reservations(resources, daterange, reservations='*'):
    query = Session().query(Reservation)
    query = query.filter(Reservation.resource.in_(resources.keys()))
    query = query.filter(Reservation.created > daterange[0])
    query = query.filter(Reservation.created <= daterange[1])
    query = query.order_by(desc(Reservation.created))

    if reservations != '*':
        query = query.filter(Reservation.token.in_(reservations))

    result = utils.OrderedDict()
    for reservation in query.all():
        if reservation.token in result:
            result[reservation.token].append(reservation)
        else:
            result[reservation.token] = [reservation]

    return result



import json
from datetime import timedelta, datetime

from five import grok
from zope.interface import Interface

from seantis.reservation.resource import CalendarRequest, get_queries
from seantis.reservation import utils
from seantis.reservation.base import BaseView, BaseViewlet
from seantis.reservation.interfaces import IOverview, OverviewletManager


class Overviewlet(BaseViewlet):
    grok.context(Interface)
    grok.name('seantis.reservation.overviewlet')
    grok.require('zope2.View')
    grok.viewletmanager(OverviewletManager)
    grok.order(1)

    overview_id = "seantis-overview-calendar"

    _template = u"""\
        <script type="text/javascript">
            if (!this.seantis) this.seantis = {};
            if (!this.seantis.overview) this.seantis.overview = {};

            this.seantis.overview.id = '#%(id)s';
            this.seantis.overview.options= %(options)s;
        </script>
        <div id="%(id)s"></div>
    """

    def overview_url(self):
        return self.context.absolute_url_path() + '/overview'

    def calendar_options(self):

        # Put the uuidmap in the json so it can be used by overview.js
        uuidmap = self.manager.uuidmap

        options = {}
        options['events'] = {
            'url': self.overview_url(),
            'type': 'POST',
            'data': {
                'uuid': uuidmap.keys()
            },
            'className': 'seantis-overview-event'
        }
        options['uuidmap'] = uuidmap

        return json.dumps(options)

    def render(self):
        if not IOverview.providedBy(self.view):
            return ''

        if not self.manager.uuidmap:
            return ''

        return self._template % {
            "id": self.overview_id,
            "options": self.calendar_options()
        }


class Overview(BaseView, CalendarRequest):
    grok.context(Interface)
    grok.name('overview')
    grok.require('zope2.View')

    def uuids(self):
        # The uuids are transmitted by the fullcalendar call, which seems to
        # mangle the the uuid options as follows:
        uuids = self.request.get('uuid[]', [])

        if not hasattr(uuids, '__iter__'):
            uuids = [uuids]

        return uuids

    def render(self):
        result = CalendarRequest.render(self)
        return result

    def events(self, daterange=None, uuids=None):
        """ Returns the events for the overview. """

        start, end = daterange or self.range
        if not all((start, end)):
            return []

        events = []

        uuids = uuids or self.uuids()
        queries = get_queries(uuids)
        days = queries.availability_by_day(start, end, uuids)

        for day, result in days.items():

            event_start = datetime(day.year, day.month, day.day, 0, 0)
            event_end = event_start + timedelta(days=+1, microseconds=-1)

            availability, resources = result
            events.append(dict(
                start=event_start.isoformat(),
                end=event_end.isoformat(),
                title=u'',
                uuids=[utils.string_uuid(r) for r in resources],
                className=utils.event_class(availability)
            ))

        return events


class Utilsviewlet(BaseViewlet):
    grok.context(Interface)
    grok.name('seantis.reservation.utilslet')
    grok.require('zope2.View')
    grok.viewletmanager(OverviewletManager)

    grok.order(10)

    template = grok.PageTemplateFile('templates/utils.pt')

    @property
    def compare_link(self):
        return utils.compare_link(self.manager.resource_uuids)

    @property
    def monthly_report_link(self):
        return utils.monthly_report_link(
            self.context, self.request, self.manager.resource_uuids
        )

    @property
    def latest_reservations_link(self):
        return utils.latest_reservations_link(
            self.context, self.request, self.manager.resource_uuids
        )

    @property
    def export_link(self):
        return utils.export_link(
            self.context, self.request, self.manager.resource_uuids
        )



import libres
import threading
import re

from five import grok
from plone import api
from seantis.reservation import utils
from sqlalchemy import create_engine
from zope.component import getUtility
from zope.event import notify
from zope.interface import implements
from zope.interface import Interface
from zope.sqlalchemy import ZopeTransactionExtension

from seantis.reservation.events import (
    ReservationsApprovedEvent,
    ReservationsDeniedEvent,
    ReservationsRevokedEvent,
    ReservationsConfirmedEvent,
    ReservationTimeChangedEvent
)


def get_postgres_version(dsn):
    """ Returns the postgres version in a tuple with the first value being
    the major version, the second being the minor version.

    Uses it's own connection to be independent from any session.

    """
    assert 'postgres' in dsn, "Not a postgres database"

    engine = create_engine(dsn)
    version = engine.execute('select version()').fetchone()[0]
    engine.dispose()

    version = re.findall('PostgreSQL (.*?) on', version)[0]
    return map(int, version.split('.'))[:2]


def assert_dsn(dsn):
    assert dsn, "Database connection not found (database.cfg)"

    if 'test://' in dsn:
        return dsn

    assert 'postgresql+psycopg2' in dsn, \
        "Only PostgreSQL combined with psycopg2 is supported"

    major, minor = get_postgres_version(dsn)

    assert (major >= 9 and minor >= 1) or (major >= 10), \
        "PostgreSQL 9.1+ is required. Your version is %i.%i" % (major, minor)

    return dsn


class ILibresUtility(Interface):
    """ Global access to libres. """

    def scheduler(name, timezone):
        pass


#
# Compatibility Shims for Libres
#
def Session():

    # LIBRES there should be a more obvious way to do this
    scheduler = getUtility(ILibresUtility).scheduler('maintenance', 'UTC')
    return scheduler.session


def db():
    scheduler = getUtility(ILibresUtility).scheduler('maintenance', 'UTC')
    return scheduler.queries
#
#
#


class CustomScheduler(libres.db.scheduler.Scheduler):
    """ Builds on the Libres scheduler to include functions that don't fit
    the scope of Libres.

    """

    def revoke_reservation(self, token, reason, id=None, send_email=True):
        """ Revoke a reservation and inform the user of that."""

        reason = reason or u''

        # sending the email first is no problem as it won't work if
        # an exception triggers later in the request

        reservations = self.reservations_by_token(token, id).all()
        notify(ReservationsRevokedEvent(
            reservations, utils.get_current_site_language(), reason, send_email
        ))

        self.remove_reservation(token, id)

    def change_reservation_time(
        self, token, id, new_start, new_end, send_email=True, reason=None
    ):
        from libres.modules import events

        def trigger_email(context, reservation, old_time, new_time):
            notify(ReservationTimeChangedEvent(
                reservation=reservation,
                language=utils.get_current_site_language(),
                old_time=old_time,
                new_time=new_time,
                send_email=send_email,
                reason=reason or u'',
            ))

        if send_email:
            events.on_reservation_time_changed.append(trigger_email)

        try:
            super(CustomScheduler, self).change_reservation_time(
                token, id, new_start, new_end
            )
        finally:
            if send_email:
                events.on_reservation_time_changed.remove(trigger_email)


class LibresUtility(grok.GlobalUtility):

    implements(ILibresUtility)

    def __init__(self):
        self.reset()
        self.setup_event_translation()

    def setup_event_translation(self):
        """ Libres events are different from the old seantis.reservation
        events. To ease the migration to libres, these events are translated
        here, where possible.

        """
        from libres.modules import events

        translated_events = (
            'on_reservations_approved',
            'on_reservations_denied',
            'on_reservations_confirmed'
        )

        for event in translated_events:
            libres_event = getattr(events, event)
            translation = getattr(self, event)

            if translation not in libres_event:
                libres_event.append(translation)

    def on_reservations_approved(self, context, reservations):
        notify(ReservationsApprovedEvent(
            reservations, utils.get_current_site_language()
        ))

    def on_reservations_denied(self, context, reservations):
        notify(ReservationsDeniedEvent(
            reservations, utils.get_current_site_language()
        ))

    def on_reservations_confirmed(self, context, reservations, session_id):
        notify(ReservationsConfirmedEvent(
            reservations, utils.get_current_site_language()
        ))

    def reset(self):
        self._dsn_cache = {}
        self._dsn_cache_lock = threading.Lock()

        try:
            self._default_dsn = utils.get_config('dsn')
        except utils.ConfigurationError:
            raise utils.ConfigurationError('No database configuration found.')

    def session_provider(self, context):
        return libres.context.session.SessionProvider(
            context.get_setting('dsn'),
            session_config={
                'extension': ZopeTransactionExtension()
            },
            engine_config={
                'json_serializer': utils.json_dumps,
                'json_deserializer': utils.json_loads
            }
        )

    def uuid_generator_factory(self, context):
        def uuid_generator(name):
            # since the name in seantis.reservation is always
            # the uuid of a resource, we don't need to generate
            # another one based on it
            return name
        return uuid_generator

    @property
    def context(self):
        site = api.portal.get()
        context_id = 'seantis.reservation' + '.'.join(site.getPhysicalPath())

        if not libres.registry.is_existing_context(context_id):
            context = libres.registry.register_context(context_id)
            context.set_service('session_provider', self.session_provider)
            context.set_setting('dsn', self.get_dsn(site))
            context.set_service('uuid_generator', self.uuid_generator_factory)
        else:
            context = libres.registry.get_context(context_id)

        return context

    def scheduler(self, name, timezone):
        return CustomScheduler(self.context, name, timezone)

    def get_dsn(self, site):
        """ Returns the DSN for the given site. Will look for those dsns
        in the zope.conf which is preferrably modified using buildout's
        <product-config> construct. See database.cfg.example for more info.

        """
        site_id = site and site.id or '__no_site__'

        if site_id not in self._dsn_cache:
            specific = utils.get_config('dsn-%s' % site_id)

            assert (specific or self._default_dsn), \
                "no dsn config found, did you define a dsn in database.cfg?"

            dsn = (specific or self._default_dsn).replace('{*}', site_id)

            self._dsn_cache_lock.acquire()
            try:
                self._dsn_cache[site_id] = assert_dsn(dsn)
            finally:
                self._dsn_cache_lock.release()

        return self._dsn_cache[site_id]



from datetime import datetime
from pytz import timezone
from seantis.reservation.tests import IntegrationTestCase
from seantis.reservation.overview import Overview


class TestOverview(IntegrationTestCase):

    def test_overview(self):
        self.login_manager()

        r1 = self.create_resource()
        r2 = self.create_resource()

        scheduler = r1.scheduler()

        start = datetime(2015, 1, 23, 12, 0)
        end = datetime(2015, 1, 23, 15, 0)
        scheduler.allocate((start, end), approve_manually=False)

        scheduler = r2.scheduler()

        start = datetime(2015, 1, 23, 12, 0)
        end = datetime(2015, 1, 23, 15, 0)
        scheduler.allocate((start, end), approve_manually=False)

        overview = Overview(r1, self.request())
        events = overview.events(
            daterange=(
                datetime(2015, 1, 21, tzinfo=timezone('UTC')),
                datetime(2015, 1, 24, tzinfo=timezone('UTC'))
            ),
            uuids=[r1.uuid(), r2.uuid()]
        )

        self.assertEqual(len(events), 1)
        self.assertEqual(events[0]['className'], 'event-available')

        scheduler.approve_reservations(
            scheduler.reserve(u'test@example.org', (start, end))
        )

        events = overview.events(
            daterange=(
                datetime(2015, 1, 21, tzinfo=timezone('UTC')),
                datetime(2015, 1, 24, tzinfo=timezone('UTC'))
            ),
            uuids=[r1.uuid(), r2.uuid()]
        )

        self.assertEqual(len(events), 1)
        self.assertEqual(events[0]['className'], 'event-partly-available')



import mock

from collections import namedtuple
from uuid import uuid1 as uuid
from datetime import datetime
from seantis.reservation.tests import IntegrationTestCase

from seantis.reservation.session import (
    getUtility,
    ILibresUtility
)

from seantis.reservation import Session
from libres.db.models import Allocation


def add_something(resource=None):
    resource = resource or uuid()
    allocation = Allocation(
        raster=15, resource=resource, mirror_of=resource)
    allocation.start = datetime(2011, 1, 1, 15)
    allocation.end = datetime(2011, 1, 1, 15, 59)
    allocation.group = uuid()

    Session.add(allocation)


class TestSession(IntegrationTestCase):

    @mock.patch('seantis.reservation.utils.get_config')
    def test_dsnconfig(self, get_config):
        util = getUtility(ILibresUtility)
        util._default_dsn = 'test://default'

        MockSite = namedtuple('MockSite', ['id'])

        get_config.return_value = None
        self.assertEqual(util.get_dsn(MockSite('test')), 'test://default')

        get_config.return_value = 'test://specific'
        self.assertEqual(util.get_dsn(MockSite('test2')), 'test://specific')

        get_config.return_value = 'test://{*}'
        self.assertEqual(util.get_dsn(MockSite('test3')), 'test://test3')

        util._default_dsn = 'test://{*}'
        get_config.return_value = None
        self.assertEqual(util.get_dsn(MockSite('test4')), 'test://test4')



from __future__ import absolute_import

from App.config import getConfiguration, setConfiguration
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting
from plone.app.testing import applyProfile
from plone.app.testing import quickInstallProduct
from plone.testing import z2
from Testing import ZopeTestCase
from testing.postgresql import Postgresql
from zope.configuration import xmlconfig


class SqlLayer(PloneSandboxLayer):

    default_bases = (PLONE_FIXTURE,)

    class Session(dict):
        def set(self, key, value):
            self[key] = value

    def start_postgres(self):
        self.postgres = Postgresql()
        return self.postgres.url().replace(
            'postgresql://', 'postgresql+psycopg2://')

    def stop_postgres(self):
        self.postgres.stop()

    def init_config(self, dsn):
        config = getConfiguration()
        if not hasattr(config, 'product_config'):
            config.product_config = {}

        config.product_config['seantis.reservation'] = dict(dsn=dsn)

        setConfiguration(config)

    def setUpZope(self, app, configurationContext):

        # Set up sessioning objects
        app.REQUEST['SESSION'] = self.Session()
        ZopeTestCase.utils.setupCoreSessions(app)

        self.init_config(dsn=self.start_postgres())

        import seantis.reservation
        xmlconfig.file(
            'configure.zcml',
            seantis.reservation,
            context=configurationContext
        )
        self.loadZCML(package=seantis.reservation)

    def setUpPloneSite(self, portal):

        quickInstallProduct(portal, 'plone.app.dexterity')
        quickInstallProduct(portal, 'seantis.reservation')
        applyProfile(portal, 'seantis.reservation:default')

    def tearDownZope(self, app):
        z2.uninstallProduct(app, 'seantis.reservation')
        self.stop_postgres()

SQL_FIXTURE = SqlLayer()

SQL_INTEGRATION_TESTING = IntegrationTesting(
    bases=(SQL_FIXTURE, ),
    name="SqlLayer:Integration"
)

SQL_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(SQL_FIXTURE, ),
    name="SqlLayer:Functional"
)



import logging

from alembic.migration import MigrationContext
from alembic.operations import Operations
from functools import wraps
from plone import api
from plone.dexterity.interfaces import IDexterityFTI
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from seantis.reservation import settings
from seantis.reservation import utils
from seantis.reservation.session import ILibresUtility
from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import types
from sqlalchemy.schema import Column
from zope.component import getUtility


log = logging.getLogger('seantis.reservation')


def db_upgrade(fn):

    @wraps(fn)
    def wrapper(context):
        # LIBRES this should be done by libres itself
        util = getUtility(ILibresUtility)
        dsn = util.get_dsn(utils.getSite())

        engine = create_engine(dsn, isolation_level='SERIALIZABLE')
        connection = engine.connect()
        transaction = connection.begin()
        try:
            context = MigrationContext.configure(connection)
            operations = Operations(context)

            metadata = MetaData(bind=engine)

            fn(operations, metadata)

            transaction.commit()

        except:
            transaction.rollback()
            raise

        finally:
            connection.close()

    return wrapper


def recook_js_resources(context):
    getToolByName(context, 'portal_javascripts').cookResources()


def recook_css_resources(context):
    getToolByName(context, 'portal_css').cookResources()


def add_new_email_template(context, new_template):
    # go through all email templates

    from seantis.reservation.mail_templates import templates
    template = templates[new_template]

    brains = utils.portal_type_in_site('seantis.reservation.emailtemplate')

    for tpl in (b.getObject() for b in brains):

        # and add the new email template in the correct language if available
        if template.is_translated(tpl.language):
            lang = tpl.language
        else:
            lang = 'en'

        setattr(
            tpl, '{}_subject'.format(new_template), template.get_subject(lang)
        )
        setattr(
            tpl, '{}_content'.format(new_template), template.get_body(lang)
        )


def remove_dead_resources(context):
    registries = [
        getToolByName(context, 'portal_javascripts'),
        getToolByName(context, 'portal_css')
    ]

    is_managed_resource = lambda r: '++seantis.reservation' in r.getId()

    def is_dead_resource(resource):

        if resource.isExternalResource():
            return False

        if context.restrictedTraverse(resource.getId(), False):
            return False

        return True

    for registry in registries:
        for resource in registry.getResources():
            if is_managed_resource(resource):
                if is_dead_resource(resource):
                    registry.unregisterResource(resource.getId())


def disable_upgrade():
    assert False, """ Unfortunately, you can't upgrade to this
    seantis.reservation release. Please install seantis.reservation 1.1.5 first
    and run its upgrade steps, before you install seantis.reservation 1.2.0.
    """


@db_upgrade
def upgrade_to_1001(operations, metadata):
    disable_upgrade()


@db_upgrade
def upgrade_1001_to_1002(operations, metadata):
    disable_upgrade()


@db_upgrade
def upgrade_1002_to_1003(operations, metadata):
    disable_upgrade()


def upgrade_1003_to_1004(context):
    disable_upgrade()


def upgrade_1004_to_1005(context):
    disable_upgrade()


def upgrade_1005_to_1006(context):
    disable_upgrade()


@db_upgrade
def upgrade_1007_to_1008(operations, metadata):
    disable_upgrade()


@db_upgrade
def upgrade_1008_to_1009(operations, metadata):
    disable_upgrade()


def upgrade_1009_to_1010(context):
    disable_upgrade()


def upgrade_1010_to_1011(context):
    disable_upgrade()


def upgrade_1011_to_1012(context):
    disable_upgrade()


def upgrade_1012_to_1013(context):
    disable_upgrade()


def upgrade_1013_to_1014(context):
    disable_upgrade()


def upgrade_1014_to_1015(context):
    disable_upgrade()


@db_upgrade
def upgrade_1015_to_1016(operations, metadata):
    disable_upgrade()


def upgrade_1016_to_1017(context):
    disable_upgrade()


def upgrade_1017_to_1018(context):
    disable_upgrade()


def upgrade_1018_to_1019(context):
    fti = getUtility(IDexterityFTI, name='seantis.reservation.resource')

    # keep the behaviors, only change the actions
    behaviors = fti.behaviors

    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'typeinfo'
    )

    fti.behaviors = behaviors


def upgrade_1019_to_1020(context):

    # add new registry values
    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'plone.app.registry'
    )


def upgrade_1020_to_1021(context):

    send_email_to_managers = settings.get('send_email_to_managers')

    registry = getUtility(IRegistry)
    registry.registerInterface(settings.ISeantisReservationSettings)

    if send_email_to_managers is True:
        settings.set('send_email_to_managers', 'by_path')
    elif send_email_to_managers is False:
        settings.set('send_email_to_managers', 'never')

    # ensure that the records exist now
    settings.get('manager_email')
    settings.get('send_email_to_managers')


def upgrade_1021_to_1022(context):

    setup = api.portal.get_tool('portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'browserlayer'
    )


def upgrade_1022_to_1023(context):

    setup = api.portal.get_tool('portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'sharing'
    )
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'rolemap'
    )


def upgrade_1023_to_1024(context):
    recook_css_resources(context)
    recook_js_resources(context)


def upgrade_1024_to_1025(context):
    recook_css_resources(context)


def upgrade_1025_to_1026(context):

    registry = getUtility(IRegistry)
    registry.registerInterface(settings.ISeantisReservationSettings)

    settings.set('available_threshold', 75)
    settings.set('partly_available_threshold', 1)

    # ensure that the records exist now
    settings.get('available_threshold')
    settings.get('partly_available_threshold')


def upgrade_1026_to_1027(context):

    setup = api.portal.get_tool('portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'jsregistry'
    )

    recook_js_resources(context)


def upgrade_1027_to_1028(context):

    upgrade_1026_to_1027(context)  # it's the same thing all over again
    recook_css_resources(context)


def upgrade_1028_to_1029(context):

    recook_css_resources(context)
    recook_js_resources(context)


def upgrade_1029_to_1030(context):

    add_new_email_template(context, 'reservation_time_changed')


def upgrade_1030_to_1031(context):

    recook_css_resources(context)
    recook_js_resources(context)


def upgrade_1031_to_1032(context):
    setup = getToolByName(context, 'portal_setup')
    setup.runImportStepFromProfile(
        'profile-seantis.reservation:default', 'typeinfo'
    )


@db_upgrade
def upgrade_1032_to_1033(operations, metadata):
    from libres.db.models.types import JSON, UTCDateTime

    # if the quota limit has been renamed, the migration already went
    # through on this database (sites may share databases)
    allocations_table = Table('allocations', metadata, autoload=True)

    if 'quota_limit' in allocations_table.columns:
        return

    # add user-data json field to allocations
    operations.add_column(
        'allocations',
        Column('data', JSON(), nullable=True))

    # add timezone to allocations (required)
    operations.add_column(
        'allocations',
        Column('timezone', types.String()))

    # add timezone to reservations (*not* required)
    operations.add_column(
        'reservations',
        Column('timezone', types.String(), nullable=True))

    # change type to
    try:
        operations.get_bind().execute("SET timezone='UTC'")

        for table in ('allocations', 'reserved_slots', 'reservations'):
            for column in ('created', 'modified'):
                operations.alter_column(
                    table, column, type_=UTCDateTime(timezone=False))
    finally:
        operations.get_bind().execute("RESET timezone")

    operations.execute(
        "UPDATE allocations SET timezone = 'UTC'")
    operations.execute(
        "UPDATE reservations SET timezone = 'UTC' WHERE start IS NOT NULL")

    # rename reservation_quota_limit to quota_limit
    operations.alter_column(
        'allocations', 'reservation_quota_limit',
        new_column_name='quota_limit')



from __future__ import print_function

import transaction
import random

from App.config import getConfiguration
from datetime import datetime, timedelta
from five import grok
from libres.db.models import Allocation
from libres.modules.rasterizer import VALID_RASTER
from plone.dexterity.utils import createContentInContainer
from seantis.reservation.base import BaseView
from seantis.reservation.error import (
    OverlappingAllocationError,
    ReservationError
)
from seantis.reservation.session import Session, ILibresUtility
from zope.component import getUtility
from zope.interface import Interface


class DataGeneratorView(BaseView):

    permission = 'cmf.ManagePortal'
    grok.require(permission)

    grok.context(Interface)
    grok.name('generate')

    template = grok.PageTemplateFile('templates/datagenerator.pt')

    @property
    def context(self):
        return getUtility(ILibresUtility).context

    @property
    def may_run(self):
        return getConfiguration().debug_mode

    @property
    def start(self):
        start = self.request.get('start', None)
        if start:
            return datetime.strptime(start, '%d.%m.%Y')
        else:
            return None

    @property
    def end(self):
        end = self.request.get('end', None)
        if end:
            return datetime.strptime(end, '%d.%m.%Y')
        else:
            return None

    @property
    def with_reservations(self):
        return bool(self.request.get('with_reservations', None))

    @property
    def min_duration(self):
        return int(self.request.get('min_duration', 30))

    @property
    def first_hour(self):
        return int(self.request.get('first_hour', 8))

    @property
    def last_hour(self):
        return int(self.request.get('last_hour', 18))

    def create_resource(self):
        resource = createContentInContainer(
            self.context, 'seantis.reservation.resource',
            title=u'random @ ' + datetime.today().strftime('%d.%m.%Y %H:%M')
        )
        resource.first_hour = self.first_hour
        resource.last_hour = self.last_hour
        return resource

    def generate_allocations(self, resource=None, start=None, end=None):

        today = datetime.today()

        resource = resource or self.create_resource()
        start = start or datetime(today.year, 1, 1)
        end = end or (start + timedelta(days=365))

        days = []
        for day in range(0, (end - start).days, 1):
            days.append(start + timedelta(days=day))

        scheduler = resource.scheduler()

        for day in days:

            for timespan in self.random_timespans(resource, day):

                quota = random.randrange(1, 1000)

                print('a @', timespan[0], timespan[1])

                try:
                    scheduler.allocate(
                        (timespan[0], timespan[1]),
                        raster=timespan[2],
                        partly_available=bool(random.randrange(0, 2)),
                        grouped=False,
                        quota=quota,
                        approve_manually=bool(random.randrange(0, 2))
                    )
                except OverlappingAllocationError:
                    pass

            # we must commit regularly or the postgres serial session
            # must track so many queries it goes to the barn and puts itself
            # down

            transaction.commit()

        if self.with_reservations:
            self.generate_reservations(resource, start, end)

    def generate_reservations(self, resource, start, end):
        query = Session.query(Allocation)
        query = query.filter(Allocation._start >= start)
        query = query.filter(Allocation._end <= end)
        query = query.filter(Allocation.mirror_of == resource.string_uuid())
        query = query.order_by(Allocation._start)

        email = 'generated@example.com'
        scheduler = resource.scheduler()

        allocations = query.all()
        Session.expunge_all()

        for allocation in allocations:

            if allocation.partly_available:
                start = allocation.start
                total = (allocation.end - allocation.start).seconds / 60

                if total > allocation.raster:
                    start_minute = random.randrange(
                        0, total - allocation.raster, allocation.raster
                    )
                    end_minute = random.randrange(
                        start_minute + allocation.raster, total,
                        allocation.raster
                    )
                else:
                    start_minute = 0
                    end_minute = total

                start = start + timedelta(start_minute * 60)
                end = start + timedelta(end_minute * 60)
            else:
                start, end = allocation.start, allocation.display_end

            if not allocation.approve_manually:
                limit = allocation.quota

            for i in range(0, random.randrange(0, limit + 1)):
                try:
                    print('r @', start, end)
                    token = scheduler.reserve(email, dates=(start, end))
                    if not allocation.approve_manually:
                        scheduler.approve_reservations(token)
                except ReservationError:
                    break

            Session.expire_on_commit = False
            transaction.commit()

    def random_raster(self):
        return random.choice(VALID_RASTER)

    def random_timespans(self, resource, day):

        day = day or datetime.today()

        min_minute = resource.first_hour * 60
        max_minute = resource.last_hour * 60

        timespans = []

        base = datetime(day.year, day.month, day.day)

        while True:
            raster = self.random_raster()
            offset = max(raster, self.min_duration)

            if max_minute - min_minute <= offset:
                break

            start_minute = random.randrange(
                min_minute, max_minute - offset, raster
            )
            end_minute = random.randrange(
                start_minute + offset, max_minute, raster
            )

            start = base + timedelta(seconds=start_minute * 60)
            end = base + timedelta(seconds=end_minute * 60)

            timespans.append((start, end, raster))

            min_minute = end_minute

        return timespans

    def update(self, *args, **kwargs):
        super(DataGeneratorView, self).update(*args, **kwargs)

        if self.may_run and self.request.get('generate_data'):
            self.generate_allocations(start=self.start, end=self.end)
            print("done!")
